# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: src/ray/protobuf/experimental/autoscaler.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n.src/ray/protobuf/experimental/autoscaler.proto\x12\x12ray.rpc.autoscaler\"X\n\x16\x41ntiAffinityConstraint\x12\x1d\n\nlabel_name\x18\x01 \x01(\tR\tlabelName\x12\x1f\n\x0blabel_value\x18\x02 \x01(\tR\nlabelValue\"T\n\x12\x41\x66\x66inityConstraint\x12\x1d\n\nlabel_name\x18\x01 \x01(\tR\tlabelName\x12\x1f\n\x0blabel_value\x18\x02 \x01(\tR\nlabelValue\"\xd3\x01\n\x13PlacementConstraint\x12T\n\ranti_affinity\x18\x01 \x01(\x0b\x32*.ray.rpc.autoscaler.AntiAffinityConstraintH\x00R\x0c\x61ntiAffinity\x88\x01\x01\x12G\n\x08\x61\x66\x66inity\x18\x02 \x01(\x0b\x32&.ray.rpc.autoscaler.AffinityConstraintH\x01R\x08\x61\x66\x66inity\x88\x01\x01\x42\x10\n\x0e_anti_affinityB\x0b\n\t_affinity\"\x98\x02\n\x0fResourceRequest\x12\x63\n\x10resources_bundle\x18\x01 \x03(\x0b\x32\x38.ray.rpc.autoscaler.ResourceRequest.ResourcesBundleEntryR\x0fresourcesBundle\x12\\\n\x15placement_constraints\x18\x02 \x03(\x0b\x32\'.ray.rpc.autoscaler.PlacementConstraintR\x14placementConstraints\x1a\x42\n\x14ResourcesBundleEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n\x05value\x18\x02 \x01(\x01R\x05value:\x02\x38\x01\"m\n\x16ResourceRequestByCount\x12=\n\x07request\x18\x01 \x01(\x0b\x32#.ray.rpc.autoscaler.ResourceRequestR\x07request\x12\x14\n\x05\x63ount\x18\x02 \x01(\x03R\x05\x63ount\"p\n\x13GangResourceRequest\x12?\n\x08requests\x18\x01 \x03(\x0b\x32#.ray.rpc.autoscaler.ResourceRequestR\x08requests\x12\x18\n\x07\x64\x65tails\x18\x02 \x01(\tR\x07\x64\x65tails\"h\n\x19\x43lusterResourceConstraint\x12K\n\x0bmin_bundles\x18\x01 \x03(\x0b\x32*.ray.rpc.autoscaler.ResourceRequestByCountR\nminBundles\"\xc1\x06\n\tNodeState\x12\x17\n\x07node_id\x18\x01 \x01(\x0cR\x06nodeId\x12\x1f\n\x0binstance_id\x18\x02 \x01(\tR\ninstanceId\x12+\n\x12ray_node_type_name\x18\x03 \x01(\tR\x0frayNodeTypeName\x12\x66\n\x13\x61vailable_resources\x18\x04 \x03(\x0b\x32\x35.ray.rpc.autoscaler.NodeState.AvailableResourcesEntryR\x12\x61vailableResources\x12Z\n\x0ftotal_resources\x18\x05 \x03(\x0b\x32\x31.ray.rpc.autoscaler.NodeState.TotalResourcesEntryR\x0etotalResources\x12W\n\x0e\x64ynamic_labels\x18\x06 \x03(\x0b\x32\x30.ray.rpc.autoscaler.NodeState.DynamicLabelsEntryR\rdynamicLabels\x12,\n\x12node_state_version\x18\x07 \x01(\x03R\x10nodeStateVersion\x12\x36\n\x06status\x18\x08 \x01(\x0e\x32\x1e.ray.rpc.autoscaler.NodeStatusR\x06status\x12(\n\x10idle_duration_ms\x18\t \x01(\x03R\x0eidleDurationMs\x12&\n\x0fnode_ip_address\x18\n \x01(\tR\rnodeIpAddress\x12,\n\x12instance_type_name\x18\x0b \x01(\tR\x10instanceTypeName\x1a\x45\n\x17\x41vailableResourcesEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n\x05value\x18\x02 \x01(\x01R\x05value:\x02\x38\x01\x1a\x41\n\x13TotalResourcesEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n\x05value\x18\x02 \x01(\x01R\x05value:\x02\x38\x01\x1a@\n\x12\x44ynamicLabelsEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n\x05value\x18\x02 \x01(\tR\x05value:\x02\x38\x01\"w\n\x1eGetClusterResourceStateRequest\x12U\n(last_seen_cluster_resource_state_version\x18\x01 \x01(\x03R#lastSeenClusterResourceStateVersion\"\xe0\x04\n\x14\x43lusterResourceState\x12\x43\n\x1e\x63luster_resource_state_version\x18\x01 \x01(\x03R\x1b\x63lusterResourceStateVersion\x12J\n\"last_seen_autoscaler_state_version\x18\x02 \x01(\x03R\x1elastSeenAutoscalerStateVersion\x12>\n\x0bnode_states\x18\x03 \x03(\x0b\x32\x1d.ray.rpc.autoscaler.NodeStateR\nnodeStates\x12\x66\n\x19pending_resource_requests\x18\x04 \x03(\x0b\x32*.ray.rpc.autoscaler.ResourceRequestByCountR\x17pendingResourceRequests\x12l\n\x1epending_gang_resource_requests\x18\x05 \x03(\x0b\x32\'.ray.rpc.autoscaler.GangResourceRequestR\x1bpendingGangResourceRequests\x12o\n\x1c\x63luster_resource_constraints\x18\x06 \x03(\x0b\x32-.ray.rpc.autoscaler.ClusterResourceConstraintR\x1a\x63lusterResourceConstraints\x12\x30\n\x14\x63luster_session_name\x18\x07 \x01(\tR\x12\x63lusterSessionName\"~\n\x1cGetClusterResourceStateReply\x12^\n\x16\x63luster_resource_state\x18\x01 \x01(\x0b\x32(.ray.rpc.autoscaler.ClusterResourceStateR\x14\x63lusterResourceState\"\xa8\x01\n\x16PendingInstanceRequest\x12,\n\x12instance_type_name\x18\x01 \x01(\tR\x10instanceTypeName\x12+\n\x12ray_node_type_name\x18\x02 \x01(\tR\x0frayNodeTypeName\x12\x14\n\x05\x63ount\x18\x03 \x01(\x05R\x05\x63ount\x12\x1d\n\nrequest_ts\x18\x04 \x01(\x03R\trequestTs\"\xd8\x01\n\x15\x46\x61iledInstanceRequest\x12,\n\x12instance_type_name\x18\x01 \x01(\tR\x10instanceTypeName\x12+\n\x12ray_node_type_name\x18\x02 \x01(\tR\x0frayNodeTypeName\x12\x14\n\x05\x63ount\x18\x03 \x01(\x05R\x05\x63ount\x12\x16\n\x06reason\x18\x04 \x01(\tR\x06reason\x12\x19\n\x08start_ts\x18\x05 \x01(\x03R\x07startTs\x12\x1b\n\tfailed_ts\x18\x06 \x01(\x03R\x08\x66\x61iledTs\"\xc6\x01\n\x0fPendingInstance\x12,\n\x12instance_type_name\x18\x01 \x01(\tR\x10instanceTypeName\x12+\n\x12ray_node_type_name\x18\x02 \x01(\tR\x0frayNodeTypeName\x12\x1f\n\x0binstance_id\x18\x03 \x01(\tR\ninstanceId\x12\x1d\n\nip_address\x18\x04 \x01(\tR\tipAddress\x12\x18\n\x07\x64\x65tails\x18\x05 \x01(\tR\x07\x64\x65tails\"\xa4\x06\n\x10\x41utoscalingState\x12U\n(last_seen_cluster_resource_state_version\x18\x01 \x01(\x03R#lastSeenClusterResourceStateVersion\x12\x38\n\x18\x61utoscaler_state_version\x18\x02 \x01(\x03R\x16\x61utoscalerStateVersion\x12\x66\n\x19pending_instance_requests\x18\x03 \x03(\x0b\x32*.ray.rpc.autoscaler.PendingInstanceRequestR\x17pendingInstanceRequests\x12\x65\n\x1cinfeasible_resource_requests\x18\x04 \x03(\x0b\x32#.ray.rpc.autoscaler.ResourceRequestR\x1ainfeasibleResourceRequests\x12r\n!infeasible_gang_resource_requests\x18\x05 \x03(\x0b\x32\'.ray.rpc.autoscaler.GangResourceRequestR\x1einfeasibleGangResourceRequests\x12\x84\x01\n\'infeasible_cluster_resource_constraints\x18\x06 \x03(\x0b\x32-.ray.rpc.autoscaler.ClusterResourceConstraintR$infeasibleClusterResourceConstraints\x12P\n\x11pending_instances\x18\x07 \x03(\x0b\x32#.ray.rpc.autoscaler.PendingInstanceR\x10pendingInstances\x12\x63\n\x18\x66\x61iled_instance_requests\x18\x08 \x03(\x0b\x32).ray.rpc.autoscaler.FailedInstanceRequestR\x16\x66\x61iledInstanceRequests\"r\n\x1dReportAutoscalingStateRequest\x12Q\n\x11\x61utoscaling_state\x18\x01 \x01(\x0b\x32$.ray.rpc.autoscaler.AutoscalingStateR\x10\x61utoscalingState\"\x1d\n\x1bReportAutoscalingStateReply\"\x98\x01\n\'RequestClusterResourceConstraintRequest\x12m\n\x1b\x63luster_resource_constraint\x18\x01 \x01(\x0b\x32-.ray.rpc.autoscaler.ClusterResourceConstraintR\x19\x63lusterResourceConstraint\"\'\n%RequestClusterResourceConstraintReply\"\x19\n\x17GetClusterStatusRequest\"\xca\x01\n\x15GetClusterStatusReply\x12Q\n\x11\x61utoscaling_state\x18\x01 \x01(\x0b\x32$.ray.rpc.autoscaler.AutoscalingStateR\x10\x61utoscalingState\x12^\n\x16\x63luster_resource_state\x18\x02 \x01(\x0b\x32(.ray.rpc.autoscaler.ClusterResourceStateR\x14\x63lusterResourceState*>\n\nNodeStatus\x12\x0f\n\x0bUNSPECIFIED\x10\x00\x12\x0b\n\x07RUNNING\x10\x01\x12\x08\n\x04\x44\x45\x41\x44\x10\x02\x12\x08\n\x04IDLE\x10\x03\x32\xa0\x04\n\x16\x41utoscalerStateService\x12\x7f\n\x17GetClusterResourceState\x12\x32.ray.rpc.autoscaler.GetClusterResourceStateRequest\x1a\x30.ray.rpc.autoscaler.GetClusterResourceStateReply\x12|\n\x16ReportAutoscalingState\x12\x31.ray.rpc.autoscaler.ReportAutoscalingStateRequest\x1a/.ray.rpc.autoscaler.ReportAutoscalingStateReply\x12\x9a\x01\n RequestClusterResourceConstraint\x12;.ray.rpc.autoscaler.RequestClusterResourceConstraintRequest\x1a\x39.ray.rpc.autoscaler.RequestClusterResourceConstraintReply\x12j\n\x10GetClusterStatus\x12+.ray.rpc.autoscaler.GetClusterStatusRequest\x1a).ray.rpc.autoscaler.GetClusterStatusReplyB\x03\xf8\x01\x01\x62\x06proto3')

_NODESTATUS = DESCRIPTOR.enum_types_by_name['NodeStatus']
NodeStatus = enum_type_wrapper.EnumTypeWrapper(_NODESTATUS)
UNSPECIFIED = 0
RUNNING = 1
DEAD = 2
IDLE = 3


_ANTIAFFINITYCONSTRAINT = DESCRIPTOR.message_types_by_name['AntiAffinityConstraint']
_AFFINITYCONSTRAINT = DESCRIPTOR.message_types_by_name['AffinityConstraint']
_PLACEMENTCONSTRAINT = DESCRIPTOR.message_types_by_name['PlacementConstraint']
_RESOURCEREQUEST = DESCRIPTOR.message_types_by_name['ResourceRequest']
_RESOURCEREQUEST_RESOURCESBUNDLEENTRY = _RESOURCEREQUEST.nested_types_by_name['ResourcesBundleEntry']
_RESOURCEREQUESTBYCOUNT = DESCRIPTOR.message_types_by_name['ResourceRequestByCount']
_GANGRESOURCEREQUEST = DESCRIPTOR.message_types_by_name['GangResourceRequest']
_CLUSTERRESOURCECONSTRAINT = DESCRIPTOR.message_types_by_name['ClusterResourceConstraint']
_NODESTATE = DESCRIPTOR.message_types_by_name['NodeState']
_NODESTATE_AVAILABLERESOURCESENTRY = _NODESTATE.nested_types_by_name['AvailableResourcesEntry']
_NODESTATE_TOTALRESOURCESENTRY = _NODESTATE.nested_types_by_name['TotalResourcesEntry']
_NODESTATE_DYNAMICLABELSENTRY = _NODESTATE.nested_types_by_name['DynamicLabelsEntry']
_GETCLUSTERRESOURCESTATEREQUEST = DESCRIPTOR.message_types_by_name['GetClusterResourceStateRequest']
_CLUSTERRESOURCESTATE = DESCRIPTOR.message_types_by_name['ClusterResourceState']
_GETCLUSTERRESOURCESTATEREPLY = DESCRIPTOR.message_types_by_name['GetClusterResourceStateReply']
_PENDINGINSTANCEREQUEST = DESCRIPTOR.message_types_by_name['PendingInstanceRequest']
_FAILEDINSTANCEREQUEST = DESCRIPTOR.message_types_by_name['FailedInstanceRequest']
_PENDINGINSTANCE = DESCRIPTOR.message_types_by_name['PendingInstance']
_AUTOSCALINGSTATE = DESCRIPTOR.message_types_by_name['AutoscalingState']
_REPORTAUTOSCALINGSTATEREQUEST = DESCRIPTOR.message_types_by_name['ReportAutoscalingStateRequest']
_REPORTAUTOSCALINGSTATEREPLY = DESCRIPTOR.message_types_by_name['ReportAutoscalingStateReply']
_REQUESTCLUSTERRESOURCECONSTRAINTREQUEST = DESCRIPTOR.message_types_by_name['RequestClusterResourceConstraintRequest']
_REQUESTCLUSTERRESOURCECONSTRAINTREPLY = DESCRIPTOR.message_types_by_name['RequestClusterResourceConstraintReply']
_GETCLUSTERSTATUSREQUEST = DESCRIPTOR.message_types_by_name['GetClusterStatusRequest']
_GETCLUSTERSTATUSREPLY = DESCRIPTOR.message_types_by_name['GetClusterStatusReply']
AntiAffinityConstraint = _reflection.GeneratedProtocolMessageType('AntiAffinityConstraint', (_message.Message,), {
  'DESCRIPTOR' : _ANTIAFFINITYCONSTRAINT,
  '__module__' : 'src.ray.protobuf.experimental.autoscaler_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.autoscaler.AntiAffinityConstraint)
  })
_sym_db.RegisterMessage(AntiAffinityConstraint)

AffinityConstraint = _reflection.GeneratedProtocolMessageType('AffinityConstraint', (_message.Message,), {
  'DESCRIPTOR' : _AFFINITYCONSTRAINT,
  '__module__' : 'src.ray.protobuf.experimental.autoscaler_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.autoscaler.AffinityConstraint)
  })
_sym_db.RegisterMessage(AffinityConstraint)

PlacementConstraint = _reflection.GeneratedProtocolMessageType('PlacementConstraint', (_message.Message,), {
  'DESCRIPTOR' : _PLACEMENTCONSTRAINT,
  '__module__' : 'src.ray.protobuf.experimental.autoscaler_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.autoscaler.PlacementConstraint)
  })
_sym_db.RegisterMessage(PlacementConstraint)

ResourceRequest = _reflection.GeneratedProtocolMessageType('ResourceRequest', (_message.Message,), {

  'ResourcesBundleEntry' : _reflection.GeneratedProtocolMessageType('ResourcesBundleEntry', (_message.Message,), {
    'DESCRIPTOR' : _RESOURCEREQUEST_RESOURCESBUNDLEENTRY,
    '__module__' : 'src.ray.protobuf.experimental.autoscaler_pb2'
    # @@protoc_insertion_point(class_scope:ray.rpc.autoscaler.ResourceRequest.ResourcesBundleEntry)
    })
  ,
  'DESCRIPTOR' : _RESOURCEREQUEST,
  '__module__' : 'src.ray.protobuf.experimental.autoscaler_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.autoscaler.ResourceRequest)
  })
_sym_db.RegisterMessage(ResourceRequest)
_sym_db.RegisterMessage(ResourceRequest.ResourcesBundleEntry)

ResourceRequestByCount = _reflection.GeneratedProtocolMessageType('ResourceRequestByCount', (_message.Message,), {
  'DESCRIPTOR' : _RESOURCEREQUESTBYCOUNT,
  '__module__' : 'src.ray.protobuf.experimental.autoscaler_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.autoscaler.ResourceRequestByCount)
  })
_sym_db.RegisterMessage(ResourceRequestByCount)

GangResourceRequest = _reflection.GeneratedProtocolMessageType('GangResourceRequest', (_message.Message,), {
  'DESCRIPTOR' : _GANGRESOURCEREQUEST,
  '__module__' : 'src.ray.protobuf.experimental.autoscaler_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.autoscaler.GangResourceRequest)
  })
_sym_db.RegisterMessage(GangResourceRequest)

ClusterResourceConstraint = _reflection.GeneratedProtocolMessageType('ClusterResourceConstraint', (_message.Message,), {
  'DESCRIPTOR' : _CLUSTERRESOURCECONSTRAINT,
  '__module__' : 'src.ray.protobuf.experimental.autoscaler_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.autoscaler.ClusterResourceConstraint)
  })
_sym_db.RegisterMessage(ClusterResourceConstraint)

NodeState = _reflection.GeneratedProtocolMessageType('NodeState', (_message.Message,), {

  'AvailableResourcesEntry' : _reflection.GeneratedProtocolMessageType('AvailableResourcesEntry', (_message.Message,), {
    'DESCRIPTOR' : _NODESTATE_AVAILABLERESOURCESENTRY,
    '__module__' : 'src.ray.protobuf.experimental.autoscaler_pb2'
    # @@protoc_insertion_point(class_scope:ray.rpc.autoscaler.NodeState.AvailableResourcesEntry)
    })
  ,

  'TotalResourcesEntry' : _reflection.GeneratedProtocolMessageType('TotalResourcesEntry', (_message.Message,), {
    'DESCRIPTOR' : _NODESTATE_TOTALRESOURCESENTRY,
    '__module__' : 'src.ray.protobuf.experimental.autoscaler_pb2'
    # @@protoc_insertion_point(class_scope:ray.rpc.autoscaler.NodeState.TotalResourcesEntry)
    })
  ,

  'DynamicLabelsEntry' : _reflection.GeneratedProtocolMessageType('DynamicLabelsEntry', (_message.Message,), {
    'DESCRIPTOR' : _NODESTATE_DYNAMICLABELSENTRY,
    '__module__' : 'src.ray.protobuf.experimental.autoscaler_pb2'
    # @@protoc_insertion_point(class_scope:ray.rpc.autoscaler.NodeState.DynamicLabelsEntry)
    })
  ,
  'DESCRIPTOR' : _NODESTATE,
  '__module__' : 'src.ray.protobuf.experimental.autoscaler_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.autoscaler.NodeState)
  })
_sym_db.RegisterMessage(NodeState)
_sym_db.RegisterMessage(NodeState.AvailableResourcesEntry)
_sym_db.RegisterMessage(NodeState.TotalResourcesEntry)
_sym_db.RegisterMessage(NodeState.DynamicLabelsEntry)

GetClusterResourceStateRequest = _reflection.GeneratedProtocolMessageType('GetClusterResourceStateRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETCLUSTERRESOURCESTATEREQUEST,
  '__module__' : 'src.ray.protobuf.experimental.autoscaler_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.autoscaler.GetClusterResourceStateRequest)
  })
_sym_db.RegisterMessage(GetClusterResourceStateRequest)

ClusterResourceState = _reflection.GeneratedProtocolMessageType('ClusterResourceState', (_message.Message,), {
  'DESCRIPTOR' : _CLUSTERRESOURCESTATE,
  '__module__' : 'src.ray.protobuf.experimental.autoscaler_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.autoscaler.ClusterResourceState)
  })
_sym_db.RegisterMessage(ClusterResourceState)

GetClusterResourceStateReply = _reflection.GeneratedProtocolMessageType('GetClusterResourceStateReply', (_message.Message,), {
  'DESCRIPTOR' : _GETCLUSTERRESOURCESTATEREPLY,
  '__module__' : 'src.ray.protobuf.experimental.autoscaler_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.autoscaler.GetClusterResourceStateReply)
  })
_sym_db.RegisterMessage(GetClusterResourceStateReply)

PendingInstanceRequest = _reflection.GeneratedProtocolMessageType('PendingInstanceRequest', (_message.Message,), {
  'DESCRIPTOR' : _PENDINGINSTANCEREQUEST,
  '__module__' : 'src.ray.protobuf.experimental.autoscaler_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.autoscaler.PendingInstanceRequest)
  })
_sym_db.RegisterMessage(PendingInstanceRequest)

FailedInstanceRequest = _reflection.GeneratedProtocolMessageType('FailedInstanceRequest', (_message.Message,), {
  'DESCRIPTOR' : _FAILEDINSTANCEREQUEST,
  '__module__' : 'src.ray.protobuf.experimental.autoscaler_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.autoscaler.FailedInstanceRequest)
  })
_sym_db.RegisterMessage(FailedInstanceRequest)

PendingInstance = _reflection.GeneratedProtocolMessageType('PendingInstance', (_message.Message,), {
  'DESCRIPTOR' : _PENDINGINSTANCE,
  '__module__' : 'src.ray.protobuf.experimental.autoscaler_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.autoscaler.PendingInstance)
  })
_sym_db.RegisterMessage(PendingInstance)

AutoscalingState = _reflection.GeneratedProtocolMessageType('AutoscalingState', (_message.Message,), {
  'DESCRIPTOR' : _AUTOSCALINGSTATE,
  '__module__' : 'src.ray.protobuf.experimental.autoscaler_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.autoscaler.AutoscalingState)
  })
_sym_db.RegisterMessage(AutoscalingState)

ReportAutoscalingStateRequest = _reflection.GeneratedProtocolMessageType('ReportAutoscalingStateRequest', (_message.Message,), {
  'DESCRIPTOR' : _REPORTAUTOSCALINGSTATEREQUEST,
  '__module__' : 'src.ray.protobuf.experimental.autoscaler_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.autoscaler.ReportAutoscalingStateRequest)
  })
_sym_db.RegisterMessage(ReportAutoscalingStateRequest)

ReportAutoscalingStateReply = _reflection.GeneratedProtocolMessageType('ReportAutoscalingStateReply', (_message.Message,), {
  'DESCRIPTOR' : _REPORTAUTOSCALINGSTATEREPLY,
  '__module__' : 'src.ray.protobuf.experimental.autoscaler_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.autoscaler.ReportAutoscalingStateReply)
  })
_sym_db.RegisterMessage(ReportAutoscalingStateReply)

RequestClusterResourceConstraintRequest = _reflection.GeneratedProtocolMessageType('RequestClusterResourceConstraintRequest', (_message.Message,), {
  'DESCRIPTOR' : _REQUESTCLUSTERRESOURCECONSTRAINTREQUEST,
  '__module__' : 'src.ray.protobuf.experimental.autoscaler_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.autoscaler.RequestClusterResourceConstraintRequest)
  })
_sym_db.RegisterMessage(RequestClusterResourceConstraintRequest)

RequestClusterResourceConstraintReply = _reflection.GeneratedProtocolMessageType('RequestClusterResourceConstraintReply', (_message.Message,), {
  'DESCRIPTOR' : _REQUESTCLUSTERRESOURCECONSTRAINTREPLY,
  '__module__' : 'src.ray.protobuf.experimental.autoscaler_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.autoscaler.RequestClusterResourceConstraintReply)
  })
_sym_db.RegisterMessage(RequestClusterResourceConstraintReply)

GetClusterStatusRequest = _reflection.GeneratedProtocolMessageType('GetClusterStatusRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETCLUSTERSTATUSREQUEST,
  '__module__' : 'src.ray.protobuf.experimental.autoscaler_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.autoscaler.GetClusterStatusRequest)
  })
_sym_db.RegisterMessage(GetClusterStatusRequest)

GetClusterStatusReply = _reflection.GeneratedProtocolMessageType('GetClusterStatusReply', (_message.Message,), {
  'DESCRIPTOR' : _GETCLUSTERSTATUSREPLY,
  '__module__' : 'src.ray.protobuf.experimental.autoscaler_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.autoscaler.GetClusterStatusReply)
  })
_sym_db.RegisterMessage(GetClusterStatusReply)

_AUTOSCALERSTATESERVICE = DESCRIPTOR.services_by_name['AutoscalerStateService']
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\370\001\001'
  _RESOURCEREQUEST_RESOURCESBUNDLEENTRY._options = None
  _RESOURCEREQUEST_RESOURCESBUNDLEENTRY._serialized_options = b'8\001'
  _NODESTATE_AVAILABLERESOURCESENTRY._options = None
  _NODESTATE_AVAILABLERESOURCESENTRY._serialized_options = b'8\001'
  _NODESTATE_TOTALRESOURCESENTRY._options = None
  _NODESTATE_TOTALRESOURCESENTRY._serialized_options = b'8\001'
  _NODESTATE_DYNAMICLABELSENTRY._options = None
  _NODESTATE_DYNAMICLABELSENTRY._serialized_options = b'8\001'
  _NODESTATUS._serialized_start=4743
  _NODESTATUS._serialized_end=4805
  _ANTIAFFINITYCONSTRAINT._serialized_start=70
  _ANTIAFFINITYCONSTRAINT._serialized_end=158
  _AFFINITYCONSTRAINT._serialized_start=160
  _AFFINITYCONSTRAINT._serialized_end=244
  _PLACEMENTCONSTRAINT._serialized_start=247
  _PLACEMENTCONSTRAINT._serialized_end=458
  _RESOURCEREQUEST._serialized_start=461
  _RESOURCEREQUEST._serialized_end=741
  _RESOURCEREQUEST_RESOURCESBUNDLEENTRY._serialized_start=675
  _RESOURCEREQUEST_RESOURCESBUNDLEENTRY._serialized_end=741
  _RESOURCEREQUESTBYCOUNT._serialized_start=743
  _RESOURCEREQUESTBYCOUNT._serialized_end=852
  _GANGRESOURCEREQUEST._serialized_start=854
  _GANGRESOURCEREQUEST._serialized_end=966
  _CLUSTERRESOURCECONSTRAINT._serialized_start=968
  _CLUSTERRESOURCECONSTRAINT._serialized_end=1072
  _NODESTATE._serialized_start=1075
  _NODESTATE._serialized_end=1908
  _NODESTATE_AVAILABLERESOURCESENTRY._serialized_start=1706
  _NODESTATE_AVAILABLERESOURCESENTRY._serialized_end=1775
  _NODESTATE_TOTALRESOURCESENTRY._serialized_start=1777
  _NODESTATE_TOTALRESOURCESENTRY._serialized_end=1842
  _NODESTATE_DYNAMICLABELSENTRY._serialized_start=1844
  _NODESTATE_DYNAMICLABELSENTRY._serialized_end=1908
  _GETCLUSTERRESOURCESTATEREQUEST._serialized_start=1910
  _GETCLUSTERRESOURCESTATEREQUEST._serialized_end=2029
  _CLUSTERRESOURCESTATE._serialized_start=2032
  _CLUSTERRESOURCESTATE._serialized_end=2640
  _GETCLUSTERRESOURCESTATEREPLY._serialized_start=2642
  _GETCLUSTERRESOURCESTATEREPLY._serialized_end=2768
  _PENDINGINSTANCEREQUEST._serialized_start=2771
  _PENDINGINSTANCEREQUEST._serialized_end=2939
  _FAILEDINSTANCEREQUEST._serialized_start=2942
  _FAILEDINSTANCEREQUEST._serialized_end=3158
  _PENDINGINSTANCE._serialized_start=3161
  _PENDINGINSTANCE._serialized_end=3359
  _AUTOSCALINGSTATE._serialized_start=3362
  _AUTOSCALINGSTATE._serialized_end=4166
  _REPORTAUTOSCALINGSTATEREQUEST._serialized_start=4168
  _REPORTAUTOSCALINGSTATEREQUEST._serialized_end=4282
  _REPORTAUTOSCALINGSTATEREPLY._serialized_start=4284
  _REPORTAUTOSCALINGSTATEREPLY._serialized_end=4313
  _REQUESTCLUSTERRESOURCECONSTRAINTREQUEST._serialized_start=4316
  _REQUESTCLUSTERRESOURCECONSTRAINTREQUEST._serialized_end=4468
  _REQUESTCLUSTERRESOURCECONSTRAINTREPLY._serialized_start=4470
  _REQUESTCLUSTERRESOURCECONSTRAINTREPLY._serialized_end=4509
  _GETCLUSTERSTATUSREQUEST._serialized_start=4511
  _GETCLUSTERSTATUSREQUEST._serialized_end=4536
  _GETCLUSTERSTATUSREPLY._serialized_start=4539
  _GETCLUSTERSTATUSREPLY._serialized_end=4741
  _AUTOSCALERSTATESERVICE._serialized_start=4808
  _AUTOSCALERSTATESERVICE._serialized_end=5352
# @@protoc_insertion_point(module_scope)



from ray.rllib.core.learner.learner import Learner
from ray.rllib.core.learner.learner_group import LearnerGroup
from ray.rllib.core.learner.learner_group_config import LearnerGroupConfig


__all__ = [
    "Learner",
    "LearnerGroup",
    "LearnerGroupConfig",
]



from functools import partial
import os
import pandas as pd
from typing import Optional, Union

from ray.air.result import Result
from ray.cloudpickle import cloudpickle
from ray.exceptions import RayTaskError
from ray.tune.analysis import ExperimentAnalysis
from ray.tune.error import TuneError
from ray.tune.experiment import Trial
from ray.tune.trainable.util import TrainableUtil
from ray.util import PublicAPI


@PublicAPI(stability="beta")
class ResultGrid:
    """A set of ``Result`` objects for interacting with Ray Tune results.

    You can use it to inspect the trials and obtain the best result.

    The constructor is a private API. This object can only be created as a result of
    ``Tuner.fit()``.

    Example:
    .. testcode::

        import random
        from ray import air, tune
        def random_error_trainable(config):
            if random.random() < 0.5:
                return {"loss": 0.0}
            else:
                raise ValueError("This is an error")
        tuner = tune.Tuner(
            random_error_trainable,
            run_config=air.RunConfig(name="example-experiment"),
            tune_config=tune.TuneConfig(num_samples=10),
        )
        try:
            result_grid = tuner.fit()
        except ValueError:
            pass
        for i in range(len(result_grid)):
            result = result_grid[i]
            if not result.error:
                    print(f"Trial finishes successfully with metrics"
                       f"{result.metrics}.")
            else:
                    print(f"Trial failed with error {result.error}.")

    .. testoutput::
        :hide:

        ...

    You can also use ``result_grid`` for more advanced analysis.

    >>> # Get the best result based on a particular metric.
    >>> best_result = result_grid.get_best_result( # doctest: +SKIP
    ...     metric="loss", mode="min")
    >>> # Get the best checkpoint corresponding to the best result.
    >>> best_checkpoint = best_result.checkpoint # doctest: +SKIP
    >>> # Get a dataframe for the last reported results of all of the trials
    >>> df = result_grid.get_dataframe() # doctest: +SKIP
    >>> # Get a dataframe for the minimum loss seen for each trial
    >>> df = result_grid.get_dataframe(metric="loss", mode="min") # doctest: +SKIP

    Note that trials of all statuses are included in the final result grid.
    If a trial is not in terminated state, its latest result and checkpoint as
    seen by Tune will be provided.

    See :doc:`/tune/examples/tune_analyze_results` for more usage examples.
    """

    def __init__(
        self,
        experiment_analysis: ExperimentAnalysis,
    ):
        self._experiment_analysis = experiment_analysis
        self._results = [
            self._trial_to_result(trial) for trial in self._experiment_analysis.trials
        ]

    @property
    def _local_path(self) -> str:
        """Return path pointing to the experiment directory on the local disk."""
        return self._experiment_analysis._local_path

    @property
    def _remote_path(self) -> Optional[str]:
        """Return path pointing to the experiment directory on remote storage."""
        return self._experiment_analysis._remote_path

    @property
    def experiment_path(self) -> str:
        """Path pointing to the experiment directory on persistent storage.

        This can point to a remote storage location (e.g. S3) or to a local
        location (path on the head node).

        For instance, if your remote storage path is ``s3://bucket/location``,
        this will point to ``s3://bucket/location/experiment_name``.
        """
        return self._remote_path or self._local_path

    def get_best_result(
        self,
        metric: Optional[str] = None,
        mode: Optional[str] = None,
        scope: str = "last",
        filter_nan_and_inf: bool = True,
    ) -> Result:
        """Get the best result from all the trials run.

        Args:
            metric: Key for trial info to order on. Defaults to
                the metric specified in your Tuner's ``TuneConfig``.
            mode: One of [min, max]. Defaults to the mode specified
                in your Tuner's ``TuneConfig``.
            scope: One of [all, last, avg, last-5-avg, last-10-avg].
                If `scope=last`, only look at each trial's final step for
                `metric`, and compare across trials based on `mode=[min,max]`.
                If `scope=avg`, consider the simple average over all steps
                for `metric` and compare across trials based on
                `mode=[min,max]`. If `scope=last-5-avg` or `scope=last-10-avg`,
                consider the simple average over the last 5 or 10 steps for
                `metric` and compare across trials based on `mode=[min,max]`.
                If `scope=all`, find each trial's min/max score for `metric`
                based on `mode`, and compare trials based on `mode=[min,max]`.
            filter_nan_and_inf: If True (default), NaN or infinite
                values are disregarded and these trials are never selected as
                the best trial.
        """
        if len(self._experiment_analysis.trials) == 1:
            return self._trial_to_result(self._experiment_analysis.trials[0])
        if not metric and not self._experiment_analysis.default_metric:
            raise ValueError(
                "No metric is provided. Either pass in a `metric` arg to "
                "`get_best_result` or specify a metric in the "
                "`TuneConfig` of your `Tuner`."
            )
        if not mode and not self._experiment_analysis.default_mode:
            raise ValueError(
                "No mode is provided. Either pass in a `mode` arg to "
                "`get_best_result` or specify a mode in the "
                "`TuneConfig` of your `Tuner`."
            )

        best_trial = self._experiment_analysis.get_best_trial(
            metric=metric,
            mode=mode,
            scope=scope,
            filter_nan_and_inf=filter_nan_and_inf,
        )
        if not best_trial:
            error_msg = (
                "No best trial found for the given metric: "
                f"{metric or self._experiment_analysis.default_metric}. "
                "This means that no trial has reported this metric"
            )
            error_msg += (
                ", or all values reported for this metric are NaN. To not ignore NaN "
                "values, you can set the `filter_nan_and_inf` arg to False."
                if filter_nan_and_inf
                else "."
            )
            raise RuntimeError(error_msg)

        return self._trial_to_result(best_trial)

    def get_dataframe(
        self,
        filter_metric: Optional[str] = None,
        filter_mode: Optional[str] = None,
    ) -> pd.DataFrame:
        """Return dataframe of all trials with their configs and reported results.

        Per default, this returns the last reported results for each trial.

        If ``filter_metric`` and ``filter_mode`` are set, the results from each
        trial are filtered for this metric and mode. For example, if
        ``filter_metric="some_metric"`` and ``filter_mode="max"``, for each trial,
        every received result is checked, and the one where ``some_metric`` is
        maximal is returned.


        Example:

            .. testcode::

                from ray.air import session
                from ray.air.config import RunConfig
                from ray.tune import Tuner

                def training_loop_per_worker(config):
                    session.report({"accuracy": 0.8})

                result_grid = Tuner(
                    trainable=training_loop_per_worker,
                    run_config=RunConfig(name="my_tune_run")
                ).fit()

                # Get last reported results per trial
                df = result_grid.get_dataframe()

                # Get best ever reported accuracy per trial
                df = result_grid.get_dataframe(
                    filter_metric="accuracy", filter_mode="max"
                )

            .. testoutput::
                :hide:

                ...

        Args:
            filter_metric: Metric to filter best result for.
            filter_mode: If ``filter_metric`` is given, one of ``["min", "max"]``
                to specify if we should find the minimum or maximum result.

        Returns:
            Pandas DataFrame with each trial as a row and their results as columns.
        """
        return self._experiment_analysis.dataframe(
            metric=filter_metric, mode=filter_mode
        )

    def __len__(self) -> int:
        return len(self._results)

    def __getitem__(self, i: int) -> Result:
        """Returns the i'th result in the grid."""
        return self._results[i]

    @property
    def errors(self):
        """Returns the exceptions of errored trials."""
        return [result.error for result in self if result.error]

    @property
    def num_errors(self):
        """Returns the number of errored trials."""
        return len(
            [t for t in self._experiment_analysis.trials if t.status == Trial.ERROR]
        )

    @property
    def num_terminated(self):
        """Returns the number of terminated (but not errored) trials."""
        return len(
            [
                t
                for t in self._experiment_analysis.trials
                if t.status == Trial.TERMINATED
            ]
        )

    @staticmethod
    def _populate_exception(trial: Trial) -> Optional[Union[TuneError, RayTaskError]]:
        if trial.status == Trial.TERMINATED:
            return None
        if trial.pickled_error_file and os.path.exists(trial.pickled_error_file):
            with open(trial.pickled_error_file, "rb") as f:
                e = cloudpickle.load(f)
                return e
        elif trial.error_file and os.path.exists(trial.error_file):
            with open(trial.error_file, "r") as f:
                return TuneError(f.read())
        return None

    def _trial_to_result(self, trial: Trial) -> Result:
        local_to_remote_path_fn = (
            partial(
                TrainableUtil.get_remote_storage_path,
                local_path_prefix=trial.local_path,
                remote_path_prefix=trial.remote_path,
            )
            if trial.uses_cloud_checkpointing
            else None
        )
        checkpoint = trial.checkpoint.to_air_checkpoint(
            local_to_remote_path_fn,
        )
        best_checkpoints = [
            (
                checkpoint.to_air_checkpoint(local_to_remote_path_fn),
                checkpoint.metrics,
            )
            for checkpoint in trial.get_trial_checkpoints()
        ]

        result = Result(
            checkpoint=checkpoint,
            metrics=trial.last_result.copy(),
            error=self._populate_exception(trial),
            _local_path=trial.local_path,
            _remote_path=trial.remote_path,
            metrics_dataframe=self._experiment_analysis.trial_dataframes.get(
                trial.local_path
            )
            if self._experiment_analysis.trial_dataframes
            else None,
            best_checkpoints=best_checkpoints,
        )
        return result

    def __repr__(self) -> str:
        all_results_repr = [result._repr(indent=2) for result in self]
        all_results_repr = ",\n".join(all_results_repr)
        return f"ResultGrid<[\n{all_results_repr}\n]>"



#!/usr/bin/env python
import importlib
import json
import os
from pathlib import Path
import sys
import typer
from typing import Optional
import yaml

import ray
from ray.tune.resources import resources_to_json, json_to_resources
from ray.tune.tune import run_experiments
from ray.tune.schedulers import create_scheduler
from ray.rllib.utils.framework import try_import_tf, try_import_torch
from ray.rllib.common import CLIArguments as cli
from ray.rllib.common import FrameworkEnum, SupportedFileType
from ray.rllib.common import download_example_file, get_file_type


def import_backends():
    """Try to import both backends for flag checking/warnings."""
    tf1, tf, tfv = try_import_tf()
    torch, _ = try_import_torch()


# Create the "train" Typer app
train_app = typer.Typer()


def _patch_path(path: str):
    """
    Patch a path to be relative to the current working directory.

    Args:
        path: relative input path.

    Returns: Patched path.
    """
    # This script runs in the ray/rllib dir.
    rllib_dir = Path(__file__).parent
    if isinstance(path, list):
        return [_patch_path(i) for i in path]
    elif isinstance(path, dict):
        return {_patch_path(k): _patch_path(v) for k, v in path.items()}
    elif isinstance(path, str):
        if os.path.exists(path):
            return path
        else:
            abs_path = str(rllib_dir.absolute().joinpath(path))
            return abs_path if os.path.exists(abs_path) else path
    else:
        return path


def load_experiments_from_file(
    config_file: str,
    file_type: SupportedFileType,
    stop: Optional[str] = None,
    checkpoint_config: Optional[dict] = None,
) -> dict:
    """Load experiments from a file. Supports YAML and Python files.

    If you want to use a Python file, it has to have a 'config' variable
    that is an AlgorithmConfig object and - optionally - a `stop` variable defining
    the stop criteria.

    Args:
        config_file: The yaml or python file to be used as experiment definition.
            Must only contain exactly one experiment.
        file_type: One value of the `SupportedFileType` enum (yaml or python).
        stop: An optional stop json string, only used if file_type is python.
            If None (and file_type is python), will try to extract stop information
            from a defined `stop` variable in the python file, otherwise, will use {}.
        checkpoint_config: An optional checkpoint config to add to the returned
            experiments dict.

    Returns:
        The experiments dict ready to be passed into `tune.run_experiments()`.
    """

    # Yaml file.
    if file_type == SupportedFileType.yaml:
        with open(config_file) as f:
            experiments = yaml.safe_load(f)
            if stop is not None and stop != "{}":
                raise ValueError("`stop` criteria only supported for python files.")
    # Python file case (ensured by file type enum)
    else:
        module_name = os.path.basename(config_file).replace(".py", "")
        spec = importlib.util.spec_from_file_location(module_name, config_file)
        module = importlib.util.module_from_spec(spec)
        sys.modules[module_name] = module
        spec.loader.exec_module(module)

        if not hasattr(module, "config"):
            raise ValueError(
                "Your Python file must contain a 'config' variable "
                "that is an AlgorithmConfig object."
            )
        algo_config = getattr(module, "config")
        if stop is None:
            stop = getattr(module, "stop", {})
        else:
            stop = json.loads(stop)

        # Note: we do this gymnastics to support the old format that
        # "run_rllib_experiments" expects. Ideally, we'd just build the config and
        # run the algo.
        config = algo_config.to_dict()
        experiments = {
            "default": {
                "run": algo_config.__class__.__name__.replace("Config", ""),
                "env": config.get("env"),
                "config": config,
                "stop": stop,
            }
        }

    for key, val in experiments.items():
        experiments[key]["checkpoint_config"] = checkpoint_config or {}

    return experiments


@train_app.command()
def file(
    # File-based arguments.
    config_file: str = cli.ConfigFile,
    # stopping conditions
    stop: Optional[str] = cli.Stop,
    # Checkpointing
    checkpoint_freq: int = cli.CheckpointFreq,
    checkpoint_at_end: bool = cli.CheckpointAtEnd,
    keep_checkpoints_num: int = cli.KeepCheckpointsNum,
    checkpoint_score_attr: str = cli.CheckpointScoreAttr,
    # Additional config arguments used for overriding.
    v: bool = cli.V,
    vv: bool = cli.VV,
    framework: FrameworkEnum = cli.Framework,
    trace: bool = cli.Trace,
    # Ray cluster options.
    local_mode: bool = cli.LocalMode,
    ray_address: Optional[str] = cli.RayAddress,
    ray_ui: bool = cli.RayUi,
    ray_num_cpus: Optional[int] = cli.RayNumCpus,
    ray_num_gpus: Optional[int] = cli.RayNumGpus,
    ray_num_nodes: Optional[int] = cli.RayNumNodes,
    ray_object_store_memory: Optional[int] = cli.RayObjectStoreMemory,
    # Ray scheduling options.
    resume: bool = cli.Resume,
    scheduler: Optional[str] = cli.Scheduler,
    scheduler_config: str = cli.SchedulerConfig,
):
    """Train a reinforcement learning agent from file.
    The file argument is required to run this command.\n\n

    Grid search example with the RLlib CLI:\n
      rllib train file tuned_examples/ppo/cartpole-ppo.yaml\n\n

    You can also run an example from a URL with the file content:\n
      rllib train file https://raw.githubusercontent.com/ray-project/ray/\
      master/rllib/tuned_examples/ppo/cartpole-ppo.yaml
    """
    # Attempt to download the file if it's not found locally.
    config_file, temp_file = download_example_file(
        example_file=config_file, base_url=None
    )

    import_backends()
    framework = framework.value if framework else None

    checkpoint_config = {
        "checkpoint_frequency": checkpoint_freq,
        "checkpoint_at_end": checkpoint_at_end,
        "num_to_keep": keep_checkpoints_num,
        "checkpoint_score_attribute": checkpoint_score_attr,
    }

    file_type = get_file_type(config_file)

    experiments = load_experiments_from_file(
        config_file, file_type, stop, checkpoint_config
    )
    exp_name = list(experiments.keys())[0]
    algo = experiments[exp_name]["run"]

    # if we had to download the config file, remove the temp file.
    if temp_file:
        temp_file.close()

    run_rllib_experiments(
        experiments=experiments,
        v=v,
        vv=vv,
        framework=framework,
        trace=trace,
        ray_num_nodes=ray_num_nodes,
        ray_num_cpus=ray_num_cpus,
        ray_num_gpus=ray_num_gpus,
        ray_object_store_memory=ray_object_store_memory,
        ray_ui=ray_ui,
        ray_address=ray_address,
        local_mode=local_mode,
        resume=resume,
        scheduler=scheduler,
        scheduler_config=scheduler_config,
        algo=algo,
    )


@train_app.callback(invoke_without_command=True)
def run(
    # Context object for subcommands
    ctx: typer.Context,
    # Config-based arguments.
    algo: str = cli.Algo,
    env: str = cli.Env,
    config: str = cli.Config,
    stop: str = cli.Stop,
    experiment_name: str = cli.ExperimentName,
    num_samples: int = cli.NumSamples,
    checkpoint_freq: int = cli.CheckpointFreq,
    checkpoint_at_end: bool = cli.CheckpointAtEnd,
    local_dir: str = cli.LocalDir,
    restore: str = cli.Restore,
    resources_per_trial: str = cli.ResourcesPerTrial,
    keep_checkpoints_num: int = cli.KeepCheckpointsNum,
    checkpoint_score_attr: str = cli.CheckpointScoreAttr,
    upload_dir: str = cli.UploadDir,
    # Additional config arguments used for overriding.
    v: bool = cli.V,
    vv: bool = cli.VV,
    framework: FrameworkEnum = cli.Framework,
    trace: bool = cli.Trace,
    # Ray cluster options.
    local_mode: bool = cli.LocalMode,
    ray_address: str = cli.RayAddress,
    ray_ui: bool = cli.RayUi,
    ray_num_cpus: int = cli.RayNumCpus,
    ray_num_gpus: int = cli.RayNumGpus,
    ray_num_nodes: int = cli.RayNumNodes,
    ray_object_store_memory: int = cli.RayObjectStoreMemory,
    # Ray scheduling options.
    resume: bool = cli.Resume,
    scheduler: str = cli.Scheduler,
    scheduler_config: str = cli.SchedulerConfig,
):
    """Train a reinforcement learning agent from command line options.
    The options --env and --algo are required to run this command.

    Training example via RLlib CLI:\n
        rllib train --algo DQN --env CartPole-v1\n\n
    """

    # If no subcommand is specified, simply run the following lines as the
    # "rllib train" main command.
    if ctx.invoked_subcommand is None:

        # we only check for backends when actually running the command. otherwise the
        # start-up time is too slow.
        import_backends()

        framework = framework.value if framework else None

        config = json.loads(config)
        resources_per_trial = json_to_resources(resources_per_trial)

        # Load a single experiment from configuration
        experiments = {
            experiment_name: {  # i.e. log to ~/ray_results/default
                "run": algo,
                "checkpoint_config": {
                    "checkpoint_frequency": checkpoint_freq,
                    "checkpoint_at_end": checkpoint_at_end,
                    "num_to_keep": keep_checkpoints_num,
                    "checkpoint_score_attribute": checkpoint_score_attr,
                },
                "local_dir": local_dir,
                "resources_per_trial": (
                    resources_per_trial and resources_to_json(resources_per_trial)
                ),
                "stop": json.loads(stop),
                "config": dict(config, env=env),
                "restore": restore,
                "num_samples": num_samples,
                "sync_config": {
                    "upload_dir": upload_dir,
                },
            }
        }

        run_rllib_experiments(
            experiments=experiments,
            v=v,
            vv=vv,
            framework=framework,
            trace=trace,
            ray_num_nodes=ray_num_nodes,
            ray_num_cpus=ray_num_cpus,
            ray_num_gpus=ray_num_gpus,
            ray_object_store_memory=ray_object_store_memory,
            ray_ui=ray_ui,
            ray_address=ray_address,
            local_mode=local_mode,
            resume=resume,
            scheduler=scheduler,
            scheduler_config=scheduler_config,
            algo=algo,
        )


def run_rllib_experiments(
    experiments: dict,
    v: cli.V,
    vv: cli.VV,
    framework: str,
    trace: cli.Trace,
    ray_num_nodes: cli.RayNumNodes,
    ray_num_cpus: cli.RayNumCpus,
    ray_num_gpus: cli.RayNumGpus,
    ray_object_store_memory: cli.RayObjectStoreMemory,
    ray_ui: cli.RayUi,
    ray_address: cli.RayAddress,
    local_mode: cli.LocalMode,
    resume: cli.Resume,
    scheduler: cli.Scheduler,
    scheduler_config: cli.SchedulerConfig,
    algo: cli.Algo,
):
    """Main training function for the RLlib CLI, whether you've loaded your
    experiments from a config file or from command line options."""

    # Override experiment data with command line arguments.
    verbose = 1
    for exp in experiments.values():
        # Bazel makes it hard to find files specified in `args` (and `data`).
        # Look for them here.
        # NOTE: Some of our yaml files don't have a `config` section.
        input_ = exp.get("config", {}).get("input")
        if input_ and input_ != "sampler":
            exp["config"]["input"] = _patch_path(input_)

        if not exp.get("env") and not exp.get("config", {}).get("env"):
            raise ValueError(
                "You either need to provide an --env argument (e.g. 'CartPole-v1') "
                "or pass an `env` key with a valid environment to your `config`"
                "argument."
            )
        elif framework is not None:
            exp["config"]["framework"] = framework
        if trace:
            if exp["config"]["framework"] not in ["tf2"]:
                raise ValueError("Must enable framework=tf2 to enable eager tracing.")
            exp["config"]["eager_tracing"] = True
        if v:
            exp["config"]["log_level"] = "INFO"
            verbose = 3  # Print details on trial result
        if vv:
            exp["config"]["log_level"] = "DEBUG"
            verbose = 3  # Print details on trial result

    # Initialize the Ray cluster with the specified options.
    if ray_num_nodes:
        # Import this only here so that train.py also works with
        # older versions (and user doesn't use `--ray-num-nodes`).
        from ray.cluster_utils import Cluster

        cluster = Cluster()
        for _ in range(ray_num_nodes):
            cluster.add_node(
                num_cpus=ray_num_cpus or 1,
                num_gpus=ray_num_gpus or 0,
                object_store_memory=ray_object_store_memory,
            )
        ray.init(address=cluster.address)
    else:
        ray.init(
            include_dashboard=ray_ui,
            address=ray_address,
            object_store_memory=ray_object_store_memory,
            num_cpus=ray_num_cpus,
            num_gpus=ray_num_gpus,
            local_mode=local_mode,
        )

    # Run the Tune experiment and return the trials.
    scheduler_config = json.loads(scheduler_config)
    trials = run_experiments(
        experiments,
        scheduler=create_scheduler(scheduler, **scheduler_config),
        resume=resume,
        verbose=verbose,
        concurrent=True,
    )
    ray.shutdown()

    checkpoints = []
    for trial in trials:
        if trial.checkpoint.dir_or_data:
            checkpoints.append(trial.checkpoint.dir_or_data)

    if checkpoints:
        from rich import print
        from rich.panel import Panel

        print("\nYour training finished.")

        print("Best available checkpoint for each trial:")
        for cp in checkpoints:
            print(f"  {cp}")

        print(
            "\nYou can now evaluate your trained algorithm from any "
            "checkpoint, e.g. by running:"
        )
        print(Panel(f"[green]  rllib evaluate {checkpoints[0]} --algo {algo}"))


def main():
    """Run the CLI."""
    train_app()


if __name__ == "__main__":
    main()



import aiorwlock
import asyncio
from importlib import import_module
import inspect
import logging
import os
import pickle
import time
from typing import Any, AsyncGenerator, Callable, Optional, Tuple, Dict
import traceback

import starlette.responses
from starlette.requests import Request
from starlette.types import Message, Receive, Scope, Send

import ray
from ray import cloudpickle
from ray.actor import ActorClass, ActorHandle
from ray.remote_function import RemoteFunction
from ray._private.async_compat import sync_to_async
from ray._private.utils import get_or_create_event_loop

from ray.serve import metrics
from ray.serve._private.common import (
    CONTROL_PLANE_CONCURRENCY_GROUP,
    ReplicaTag,
    ServeComponentType,
)
from ray.serve.config import DeploymentConfig
from ray.serve._private.constants import (
    HEALTH_CHECK_METHOD,
    RECONFIGURE_METHOD,
    DEFAULT_LATENCY_BUCKET_MS,
    SERVE_LOGGER_NAME,
    SERVE_NAMESPACE,
    RAY_SERVE_GAUGE_METRIC_SET_PERIOD_S,
)
from ray.serve.deployment import Deployment
from ray.serve.exceptions import RayServeException
from ray.serve._private.http_util import (
    make_buffered_asgi_receive,
    ASGIAppReplicaWrapper,
    ASGIReceiveProxy,
    ASGIMessageQueue,
    BufferedASGISender,
    HTTPRequestWrapper,
    RawASGIResponse,
    Response,
)
from ray.serve._private.logging_utils import (
    access_log_msg,
    configure_component_logger,
    get_component_logger_file_path,
)
from ray.serve._private.router import RequestMetadata
from ray.serve._private.utils import (
    parse_import_path,
    wrap_to_ray_error,
    merge_dict,
    MetricsPusher,
)
from ray.serve._private.version import DeploymentVersion


logger = logging.getLogger(SERVE_LOGGER_NAME)


def _format_replica_actor_name(deployment_name: str):
    return f"ServeReplica:{deployment_name}"


def create_replica_wrapper(name: str):
    """Creates a replica class wrapping the provided function or class.

    This approach is picked over inheritance to avoid conflict between user
    provided class and the RayServeReplica class.
    """

    # TODO(architkulkarni): Add type hints after upgrading cloudpickle
    class RayServeWrappedReplica(object):
        async def __init__(
            self,
            deployment_name,
            replica_tag,
            serialized_deployment_def: bytes,
            serialized_init_args: bytes,
            serialized_init_kwargs: bytes,
            deployment_config_proto_bytes: bytes,
            version: DeploymentVersion,
            controller_name: str,
            detached: bool,
            app_name: str = None,
        ):
            self._replica_tag = replica_tag
            configure_component_logger(
                component_type=ServeComponentType.DEPLOYMENT,
                component_name=deployment_name,
                component_id=replica_tag,
            )

            self._event_loop = get_or_create_event_loop()

            deployment_def = cloudpickle.loads(serialized_deployment_def)

            if isinstance(deployment_def, str):
                import_path = deployment_def
                module_name, attr_name = parse_import_path(import_path)
                deployment_def = getattr(import_module(module_name), attr_name)
                # For ray or serve decorated class or function, strip to return
                # original body
                if isinstance(deployment_def, RemoteFunction):
                    deployment_def = deployment_def._function
                elif isinstance(deployment_def, ActorClass):
                    deployment_def = deployment_def.__ray_metadata__.modified_class
                elif isinstance(deployment_def, Deployment):
                    logger.warning(
                        f'The import path "{import_path}" contains a '
                        "decorated Serve deployment. The decorator's settings "
                        "are ignored when deploying via import path."
                    )
                    deployment_def = deployment_def.func_or_class

            init_args = cloudpickle.loads(serialized_init_args)
            init_kwargs = cloudpickle.loads(serialized_init_kwargs)

            deployment_config = DeploymentConfig.from_proto_bytes(
                deployment_config_proto_bytes
            )

            if inspect.isfunction(deployment_def):
                is_function = True
            elif inspect.isclass(deployment_def):
                is_function = False
            else:
                assert False, (
                    "deployment_def must be function, class, or "
                    "corresponding import path. Instead, it's type was "
                    f"{type(deployment_def)}."
                )

            # Set the controller name so that serve.connect() in the user's
            # code will connect to the instance that this deployment is running
            # in.
            ray.serve.context._set_internal_replica_context(
                deployment_name,
                replica_tag,
                controller_name,
                servable_object=None,
                app_name=app_name,
            )

            assert controller_name, "Must provide a valid controller_name"

            controller_handle = ray.get_actor(
                controller_name, namespace=SERVE_NAMESPACE
            )

            # Indicates whether the replica has finished initializing.
            self._initialized = False

            # This closure initializes user code and finalizes replica
            # startup. By splitting the initialization step like this,
            # we can already access this actor before the user code
            # has finished initializing.
            # The supervising state manager can then wait
            # for allocation of this replica by using the `is_allocated`
            # method. After that, it calls `reconfigure` to trigger
            # user code initialization.
            async def initialize_replica():
                if is_function:
                    _callable = deployment_def
                else:
                    # This allows deployments to define an async __init__
                    # method (mostly used for testing).
                    _callable = deployment_def.__new__(deployment_def)
                    await sync_to_async(_callable.__init__)(*init_args, **init_kwargs)

                    if isinstance(_callable, ASGIAppReplicaWrapper):
                        await _callable._run_asgi_lifespan_startup()

                # Setting the context again to update the servable_object.
                ray.serve.context._set_internal_replica_context(
                    deployment_name,
                    replica_tag,
                    controller_name,
                    servable_object=_callable,
                    app_name=app_name,
                )

                self.replica = RayServeReplica(
                    _callable,
                    deployment_name,
                    replica_tag,
                    deployment_config.autoscaling_config,
                    version,
                    is_function,
                    controller_handle,
                    app_name,
                )
                self._initialized = True

            # Is it fine that replica is None here?
            # Should we add a check in all methods that use self.replica
            # or, alternatively, create an async get_replica() method?
            self.replica = None
            self._initialize_replica = initialize_replica

            # Used to guard `initialize_replica` so that it isn't called twice.
            self._replica_init_lock = asyncio.Lock()

        @ray.method(concurrency_group=CONTROL_PLANE_CONCURRENCY_GROUP)
        def get_num_ongoing_requests(self) -> int:
            """Fetch the number of ongoing requests at this replica (queue length).

            This runs on a separate thread (using a Ray concurrency group) so it will
            not be blocked by user code.
            """
            return self.replica.get_num_pending_and_running_requests()

        @ray.method(num_returns=2)
        async def handle_request(
            self,
            pickled_request_metadata: bytes,
            *request_args,
            **request_kwargs,
        ) -> Tuple[bytes, Any]:

            request_metadata = pickle.loads(pickled_request_metadata)
            if request_metadata.is_http_request:
                # The sole argument passed from `http_proxy.py` is the ASGI scope.
                assert len(request_args) == 1
                request: HTTPRequestWrapper = pickle.loads(request_args[0])

                scope = request.scope
                buffered_send = BufferedASGISender()
                buffered_receive = make_buffered_asgi_receive(request.body)
                request_args = (scope, buffered_receive, buffered_send)

            result = await self.replica.handle_request(
                request_metadata, request_args, request_kwargs
            )

            if request_metadata.is_http_request:
                result = buffered_send.build_asgi_response()

            # Returns a small object for router to track request status.
            return b"", result

        async def handle_request_streaming(
            self,
            pickled_request_metadata: bytes,
            pickled_asgi_scope: bytes,
            http_proxy_handle: ActorHandle,
        ) -> AsyncGenerator[Message, None]:
            """Handle a request and stream the results to the caller.

            This is used by the HTTP proxy for experimental StreamingResponse support.

            This generator yields ASGI-compliant messages sent via an ASGI send
            interface. This allows us to return the messages back to the HTTP proxy as
            they're sent by user code (e.g., the FastAPI wrapper).
            """
            request_metadata = pickle.loads(pickled_request_metadata)
            if not request_metadata.is_http_request:
                raise NotImplementedError(
                    "Only HTTP requests are currently supported over streaming."
                )

            receiver_task = None
            handle_request_task = None
            wait_for_message_task = None
            try:
                receiver = ASGIReceiveProxy(
                    request_metadata.request_id, http_proxy_handle
                )
                receiver_task = self._event_loop.create_task(
                    receiver.fetch_until_disconnect()
                )

                scope = pickle.loads(pickled_asgi_scope)
                asgi_queue_send = ASGIMessageQueue()
                request_args = (scope, receiver, asgi_queue_send)
                request_kwargs = {}

                # Handle the request in a background asyncio.Task. It's expected that
                # this task will use the provided ASGI send interface to send its HTTP
                # the response. We will poll for the sent messages and yield them back
                # to the caller.
                handle_request_task = self._event_loop.create_task(
                    self.replica.handle_request(
                        request_metadata, request_args, request_kwargs
                    )
                )

                while True:
                    wait_for_message_task = self._event_loop.create_task(
                        asgi_queue_send.wait_for_message()
                    )
                    done, _ = await asyncio.wait(
                        [handle_request_task, wait_for_message_task],
                        return_when=asyncio.FIRST_COMPLETED,
                    )
                    # Consume and yield all available messages in the queue.
                    # The messages are batched into a list to avoid unnecessary RPCs and
                    # we use vanilla pickle because it's faster than cloudpickle and we
                    # know it's safe for these messages containing primitive types.
                    yield pickle.dumps(asgi_queue_send.get_messages_nowait())

                    # Exit once `handle_request` has finished. In this case, all
                    # messages must have already been sent.
                    if handle_request_task in done:
                        break

                e = handle_request_task.exception()
                if e is not None:
                    raise e from None
            finally:
                if receiver_task is not None:
                    receiver_task.cancel()

                if handle_request_task is not None and not handle_request_task.done():
                    handle_request_task.cancel()

                if (
                    wait_for_message_task is not None
                    and not wait_for_message_task.done()
                ):
                    wait_for_message_task.cancel()

        async def handle_request_from_java(
            self,
            proto_request_metadata: bytes,
            *request_args,
            **request_kwargs,
        ) -> Any:
            from ray.serve.generated.serve_pb2 import (
                RequestMetadata as RequestMetadataProto,
            )

            proto = RequestMetadataProto.FromString(proto_request_metadata)
            request_metadata: RequestMetadata = RequestMetadata(
                proto.request_id, proto.endpoint, call_method=proto.call_method
            )
            request_args = request_args[0]
            return await self.replica.handle_request(
                request_metadata, request_args, request_kwargs
            )

        async def is_allocated(self) -> str:
            """poke the replica to check whether it's alive.

            When calling this method on an ActorHandle, it will complete as
            soon as the actor has started running. We use this mechanism to
            detect when a replica has been allocated a worker slot.
            At this time, the replica can transition from PENDING_ALLOCATION
            to PENDING_INITIALIZATION startup state.

            Returns:
                The PID, actor ID, node ID, node IP, and log filepath id of the replica.
            """

            return (
                os.getpid(),
                ray.get_runtime_context().get_actor_id(),
                ray.get_runtime_context().get_worker_id(),
                ray.get_runtime_context().get_node_id(),
                ray.util.get_node_ip_address(),
                get_component_logger_file_path(),
            )

        async def initialize_and_get_metadata(
            self,
            deployment_config: DeploymentConfig = None,
            _after: Optional[Any] = None,
        ) -> Tuple[DeploymentConfig, DeploymentVersion]:
            # Unused `_after` argument is for scheduling: passing an ObjectRef
            # allows delaying this call until after the `_after` call has returned.
            try:
                # Ensure that initialization is only performed once.
                # When controller restarts, it will call this method again.
                async with self._replica_init_lock:
                    if not self._initialized:
                        await self._initialize_replica()
                    if deployment_config:
                        await self.replica.update_user_config(
                            deployment_config.user_config
                        )

                # A new replica should not be considered healthy until it passes
                # an initial health check. If an initial health check fails,
                # consider it an initialization failure.
                await self.check_health()
                return await self._get_metadata()
            except Exception:
                raise RuntimeError(traceback.format_exc()) from None

        async def reconfigure(
            self,
            deployment_config: DeploymentConfig,
        ) -> Tuple[DeploymentConfig, DeploymentVersion]:
            try:
                await self.replica.reconfigure(deployment_config)
                return await self._get_metadata()
            except Exception:
                raise RuntimeError(traceback.format_exc()) from None

        async def _get_metadata(
            self,
        ) -> Tuple[DeploymentConfig, DeploymentVersion]:
            return self.replica.version.deployment_config, self.replica.version

        async def prepare_for_shutdown(self):
            if self.replica is not None:
                return await self.replica.prepare_for_shutdown()

        @ray.method(concurrency_group=CONTROL_PLANE_CONCURRENCY_GROUP)
        async def check_health(self):
            await self.replica.check_health()

    # Dynamically create a new class with custom name here so Ray picks it up
    # correctly in actor metadata table and observability stack.
    return type(
        _format_replica_actor_name(name),
        (RayServeWrappedReplica,),
        dict(RayServeWrappedReplica.__dict__),
    )


class RayServeReplica:
    """Handles requests with the provided callable."""

    def __init__(
        self,
        _callable: Callable,
        deployment_name: str,
        replica_tag: ReplicaTag,
        autoscaling_config: Any,
        version: DeploymentVersion,
        is_function: bool,
        controller_handle: ActorHandle,
        app_name: str,
    ) -> None:
        self.deployment_name = deployment_name
        self.replica_tag = replica_tag
        self.callable = _callable
        self.is_function = is_function
        self.version = version
        self.deployment_config: DeploymentConfig = version.deployment_config
        self.rwlock = aiorwlock.RWLock()
        self.delete_lock = asyncio.Lock()
        self.app_name = app_name

        user_health_check = getattr(_callable, HEALTH_CHECK_METHOD, None)
        if not callable(user_health_check):

            def user_health_check():
                pass

        self.user_health_check = sync_to_async(user_health_check)

        self.request_counter = metrics.Counter(
            "serve_deployment_request_counter",
            description=(
                "The number of queries that have been processed in this replica."
            ),
            tag_keys=("route",),
        )

        self.error_counter = metrics.Counter(
            "serve_deployment_error_counter",
            description=(
                "The number of exceptions that have occurred in this replica."
            ),
            tag_keys=("route",),
        )

        self.restart_counter = metrics.Counter(
            "serve_deployment_replica_starts",
            description=(
                "The number of times this replica has been restarted due to failure."
            ),
        )

        self.processing_latency_tracker = metrics.Histogram(
            "serve_deployment_processing_latency_ms",
            description="The latency for queries to be processed.",
            boundaries=DEFAULT_LATENCY_BUCKET_MS,
            tag_keys=("route",),
        )

        self.num_processing_items = metrics.Gauge(
            "serve_replica_processing_queries",
            description="The current number of queries being processed.",
        )

        self.num_pending_items = metrics.Gauge(
            "serve_replica_pending_queries",
            description="The current number of pending queries.",
        )

        self.restart_counter.inc()

        self.metrics_pusher = self.metrics_pusher = MetricsPusher()
        if autoscaling_config:
            process_remote_func = controller_handle.record_autoscaling_metrics.remote
            config = autoscaling_config
            self.metrics_pusher.register_task(
                self.collect_autoscaling_metrics,
                config.metrics_interval_s,
                process_remote_func,
            )

        self.metrics_pusher.register_task(
            self._set_replica_requests_metrics,
            RAY_SERVE_GAUGE_METRIC_SET_PERIOD_S,
        )
        self.metrics_pusher.start()

    def _set_replica_requests_metrics(self):
        self.num_processing_items.set(self.get_num_running_requests())
        self.num_pending_items.set(self.get_num_pending_requests())

    async def check_health(self):
        await self.user_health_check()

    def _get_handle_request_stats(self) -> Optional[Dict[str, int]]:
        replica_actor_name = _format_replica_actor_name(self.deployment_name)
        actor_stats = ray.runtime_context.get_runtime_context()._get_actor_call_stats()
        method_stats = actor_stats.get(f"{replica_actor_name}.handle_request")
        streaming_method_stats = actor_stats.get(
            f"{replica_actor_name}.handle_request_streaming"
        )
        method_stats_java = actor_stats.get(
            f"{replica_actor_name}.handle_request_from_java"
        )
        return merge_dict(
            merge_dict(method_stats, streaming_method_stats), method_stats_java
        )

    def get_num_running_requests(self) -> int:
        stats = self._get_handle_request_stats() or {}
        return stats.get("running", 0)

    def get_num_pending_requests(self) -> int:
        stats = self._get_handle_request_stats() or {}
        return stats.get("pending", 0)

    def get_num_pending_and_running_requests(self) -> int:
        stats = self._get_handle_request_stats() or {}
        return stats.get("pending", 0) + stats.get("running", 0)

    def collect_autoscaling_metrics(self):
        return {self.replica_tag: self.get_num_pending_and_running_requests()}

    def get_runner_method(self, request_metadata: RequestMetadata) -> Callable:
        method_name = request_metadata.call_method
        if not hasattr(self.callable, method_name):
            # Filter to methods that don't start with '__' prefix.
            def callable_method_filter(attr):
                if attr.startswith("__"):
                    return False
                elif not callable(getattr(self.callable, attr)):
                    return False

                return True

            methods = list(filter(callable_method_filter, dir(self.callable)))
            raise RayServeException(
                f"Tried to call a method '{method_name}' "
                "that does not exist. Available methods: "
                f"{methods}."
            )
        if self.is_function:
            return self.callable

        return getattr(self.callable, method_name)

    async def send_user_result_over_asgi(
        self, result: Any, scope: Scope, receive: Receive, send: Send
    ):
        """Handle the result from user code and send it over the ASGI interface.

        If the result is already a Response type, it is sent directly. Otherwise, it
        is converted to a custom Response type that handles serialization for
        common Python objects.
        """
        if not isinstance(result, (starlette.responses.Response, RawASGIResponse)):
            await Response(result).send(scope, receive, send)
        else:
            await result(scope, receive, send)

    async def invoke_single(
        self,
        request_metadata: RequestMetadata,
        request_args: Tuple[Any],
        request_kwargs: Dict[str, Any],
    ) -> Tuple[Any, bool]:
        """Executes the provided request on this replica.

        Returns the user-provided output and a boolean indicating if the
        request succeeded (user code didn't raise an exception).
        """
        logger.info(
            f"Started executing request {request_metadata.request_id}",
            extra={"log_to_stderr": False},
        )

        if request_metadata.is_http_request:
            # For HTTP requests we always expect (scope, receive, send) as args.
            assert len(request_args) == 3
            scope, receive, send = request_args

            if isinstance(self.callable, ASGIAppReplicaWrapper):
                request_args = (scope, receive, send)
            else:
                request_args = (Request(scope, receive, send),)

        method_to_call = None
        success = True
        try:
            runner_method = self.get_runner_method(request_metadata)
            method_to_call = sync_to_async(runner_method)
            result = None

            # Edge case to support empty HTTP handlers: don't pass the Request
            # argument if the callable has no parameters.
            if (
                request_metadata.is_http_request
                and len(inspect.signature(runner_method).parameters) == 0
            ):
                request_args, request_kwargs = tuple(), {}

            result = await method_to_call(*request_args, **request_kwargs)

        except Exception as e:
            logger.exception(f"Request failed due to {type(e).__name__}:")
            success = False

            # If the debugger is enabled, drop into the remote pdb here.
            if ray.util.pdb._is_ray_debugger_enabled():
                ray.util.pdb._post_mortem()

            function_name = "unknown"
            if method_to_call is not None:
                function_name = method_to_call.__name__
            result = wrap_to_ray_error(function_name, e)
            if request_metadata.is_http_request:
                error_message = f"Unexpected error, traceback: {result}."
                result = starlette.responses.Response(error_message, status_code=500)

        if request_metadata.is_http_request and not isinstance(
            self.callable, ASGIAppReplicaWrapper
        ):
            # For the FastAPI codepath, the response has already been sent over the ASGI
            # interface, but for the vanilla deployment codepath we need to send it.
            await self.send_user_result_over_asgi(result, scope, receive, send)

        if success:
            self.request_counter.inc(tags={"route": request_metadata.route})
        else:
            self.error_counter.inc(tags={"route": request_metadata.route})

        return result, success

    async def reconfigure(self, deployment_config: DeploymentConfig):
        old_user_config = self.deployment_config.user_config
        self.deployment_config = deployment_config
        self.version = DeploymentVersion.from_deployment_version(
            self.version, self.deployment_config
        )

        if old_user_config != deployment_config.user_config:
            await self.update_user_config(deployment_config.user_config)

    async def update_user_config(self, user_config: Any):
        async with self.rwlock.writer_lock:
            if user_config is not None:
                if self.is_function:
                    raise ValueError(
                        "deployment_def must be a class to use user_config"
                    )
                elif not hasattr(self.callable, RECONFIGURE_METHOD):
                    raise RayServeException(
                        "user_config specified but deployment "
                        + self.deployment_name
                        + " missing "
                        + RECONFIGURE_METHOD
                        + " method"
                    )
                reconfigure_method = sync_to_async(
                    getattr(self.callable, RECONFIGURE_METHOD)
                )
                await reconfigure_method(user_config)

    async def handle_request(
        self,
        request_metadata: RequestMetadata,
        request_args: Tuple[Any],
        request_kwargs: Dict[str, Any],
    ) -> Any:
        async with self.rwlock.reader_lock:
            # Set request context variables for subsequent handle so that
            # handle can pass the correct request context to subsequent replicas.
            ray.serve.context._serve_request_context.set(
                ray.serve.context.RequestContext(
                    request_metadata.route,
                    request_metadata.request_id,
                    self.app_name,
                    request_metadata.multiplexed_model_id,
                )
            )

            start_time = time.time()
            result, success = await self.invoke_single(
                request_metadata,
                request_args,
                request_kwargs,
            )
            latency_ms = (time.time() - start_time) * 1000
            self.processing_latency_tracker.observe(
                latency_ms, tags={"route": request_metadata.route}
            )
            logger.info(
                access_log_msg(
                    method=request_metadata.call_method,
                    status="OK" if success else "ERROR",
                    latency_ms=latency_ms,
                )
            )

            return result

    async def prepare_for_shutdown(self):
        """Perform graceful shutdown.

        Trigger a graceful shutdown protocol that will wait for all the queued
        tasks to be completed and return to the controller.
        """
        while True:
            # Sleep first because we want to make sure all the routers receive
            # the notification to remove this replica first.
            await asyncio.sleep(self.deployment_config.graceful_shutdown_wait_loop_s)

            num_ongoing_requests = self.get_num_pending_and_running_requests()
            if num_ongoing_requests > 0:
                logger.info(
                    "Waiting for an additional "
                    f"{self.deployment_config.graceful_shutdown_wait_loop_s}s to shut "
                    f"down because there are {num_ongoing_requests} ongoing "
                    "requests."
                )
            else:
                logger.info(
                    "Graceful shutdown complete; replica exiting.",
                    extra={"log_to_stderr": False},
                )
                break

        # Explicitly call the del method to trigger clean up.
        # We set the del method to noop after successfully calling it so the
        # destructor is called only once.
        async with self.delete_lock:
            if not hasattr(self, "callable"):
                return

            try:
                if hasattr(self.callable, "__del__"):
                    # Make sure to accept `async def __del__(self)` as well.
                    await sync_to_async(self.callable.__del__)()
                    setattr(self.callable, "__del__", lambda _: None)
            except Exception as e:
                logger.exception(f"Exception during graceful shutdown of replica: {e}")
            finally:
                if hasattr(self.callable, "__del__"):
                    del self.callable
